/*  
    NWP Dict -- NieWielki Parser Sownika PWN
    NWP Dict -- Not Worth the Pay (so it's free;) -- PWN Dictionary Parser 

    Written by Jakub Nadolny, February 2009
    jakub@nadolny.info

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// ! nie dziaa tumacz. who, who'd

#include <iostream>
#include <zlib.h> // for definitions uncompress

#include <fstream>
#include <cstdlib>
#include <cstring>
#include <sys/mman.h>
#include <fcntl.h>
#include <string.h>

// #define DEBUG 1; // debug mode

using namespace std;

const char *config_file_path = "/etc/nwpdict.conf"; // does not work yet!
const char *polang_dict_file = "/usr/share/kydpdict/polang.win";
const char *angpol_dict_file = "/usr/share/kydpdict/angpol.win";

unsigned int fix32(unsigned int x)
{
#ifndef WORDS_BIGENDIAN
        return x;
#else
        return (unsigned int)
                (((x & (unsigned int) 0x000000ffU) << 24) |
                 ((x & (unsigned int) 0x0000ff00U) << 8) |
                 ((x & (unsigned int) 0x00ff0000U) >> 8) |
                 ((x & (unsigned int) 0xff000000U) >> 24));
#endif
}

char *dict_file_name=(char *)malloc(150); // full path and name to dictionary file
string word;   // word to search for
enum action_type { toPl, toEn, listEnWords, listPlWords } action; // action to be performed

// str = s/substr1/substr2/
void replace(string *str, const string substr1, const string substr2) {
 size_t tag=str->find(substr1);
 while (tag != string::npos) {
  str->replace(tag,substr1.length(),substr2);
  tag=str->find(substr1);
 } 
}

// converts pl diactrical characters from windows1250 to iso-8859-2
void win2iso(string *str) {
 replace(str,"\xB9","");
 replace(str,"\xA5","");
 replace(str,"\x9c","");
 replace(str,"\x8c","");
 replace(str,"\x9f","");
 replace(str,"\x8f","");
}

// removes all <.*> tags
void remove_tags(string *str) {
 size_t tag=str->find_first_of("<");
 while (tag != string::npos) {
  string::iterator it1, it2;
  it1 = str->begin() + str->find_first_of("<");
  it2 = str->begin() + str->find_first_of(">")+1;
  str->erase(it1, it2);
  tag=str->find_first_of("<");
 }
}

class cDictionary {
 private:
  int inioffset,mmapbase,endbase; // czy to powinny by skadowe klasy???
  int wordCount;       // total number of words
  int index_base;      // offset for indexes in dict file
  int words_base;      // offset for words in dict file
  unsigned int *offsets; // offsets for words definitions
  unsigned int *noffsets; // sprawdzi to
  char **words;          // words from dictionary
  ifstream dict_file;
  string xml_parse(string str); 
  void print_definition(int w); // read and print definition of word number w
 public:
  cDictionary() { wordCount=0; index_base=0; words_base=0; }
  void read_bases();
  void read_offsets();
  void read_words();
  void read_definition(string word);
};

string cDictionary::xml_parse(string str) {
 replace(&str,"<P>","\n"); // change <P> to new line
 replace(&str,"&##39;","@");
 replace(&str,"&aelig;","{");
 replace(&str,"&##8747;","S");
 replace(&str,"&inodot;","I");
 replace(&str,"&rsquo;","'");
 replace(&str,"&lsquo;","'");
 replace(&str,"&quotup;","\"");
 replace(&str,"&quotlw;","\"");
 replace(&str,"&hfpause;","-");
 replace(&str,"&pause;","-");
 replace(&str,"&rarr;","zob.");
 replace(&str,"&dots;","...");
 replace(&str,"&squareb;","-");

 win2iso(&str);
 remove_tags(&str);

 return str;
}

// reads and prints definition of word number w
void cDictionary::print_definition(int w) {
 unsigned int maxlength = 100000;
 char *wordbuffer = new char[maxlength+1];

#ifdef DEBUG
 cout << "w=" << w << ", offsets[w]=" << offsets[w] << endl;
#endif
 dict_file.seekg(words_base+fix32(offsets[w]),ios::beg); 
 dict_file.read(wordbuffer,maxlength);
 
 static char *outbuffer = NULL; 

 if (outbuffer==NULL)
  outbuffer = new char [5*maxlength];     // XXX is this memleak?
 else
  memset(outbuffer,0,5*maxlength);        // 5*maxlength is arbitrary, based on zip performance
  
 int i = 2+4+6+strlen(&wordbuffer[2+4+6])+2;
 if (wordbuffer[i]<20) {                 // not a character -> must be compressed
  int unzipresult;
  uLongf destlen;
  i += wordbuffer[i]+1;
  destlen = 5*maxlength;
  unzipresult = uncompress((Bytef*)outbuffer,&destlen,(const Bytef*)&wordbuffer[i],maxlength); 
 } 

#ifdef DEBUG
 cout << endl << outbuffer << endl;
#endif
 cout << endl << xml_parse(outbuffer) << endl;
}

void cDictionary::read_bases() {
 char buffer[128];
 size_t page_size;
 unsigned int end_last_word;

 dict_file.open(dict_file_name); 

 dict_file.get(buffer, 128);

 memcpy(&wordCount, &buffer[24], 4); // should be 63378
#ifdef DEBUG 
 cout << "Number of words: " << wordCount << endl; 
#endif
 memcpy(&index_base, &buffer[28], 4); // should be 23016700
 memcpy(&words_base, &buffer[32], 4);

 offsets = new unsigned int[wordCount+1];
 noffsets = new unsigned int[wordCount+1];
 words = new char* [wordCount+1];
 words[wordCount] = 0;

 dict_file.seekg(index_base);
 dict_file.read((char *)offsets,4*wordCount); 

 for (int i=0;i<wordCount;i++)
  offsets[i] = fix32(offsets[i]) & 0x07ffffff; 

  /* read index table */
 page_size = (size_t)sysconf(_SC_PAGESIZE);
 mmapbase  = (words_base/page_size)*page_size;
 inioffset = words_base - mmapbase;

 // printf("page_size: %u, mmapbase: %u, inioffset: %u\n",page_size,mmapbase,inioffset); // jest ok!

 dict_file.seekg(0,ios::end);
 
 end_last_word = dict_file.tellg();
 endbase   = ((end_last_word/page_size)+1)*page_size;
}

// Read all words into words[]
void cDictionary::read_words() {
#ifdef DEBUG 
 cout << "Reading words... ";
#endif
 int f = open(dict_file_name,O_RDONLY);
 char *filedata = (char*)mmap(NULL, endbase, PROT_READ, MAP_PRIVATE, f, mmapbase);

 int j = 0;
  // read words from dictionary file
 if ((void *)filedata != MAP_FAILED) {
  for (int i=0;i<wordCount;i++) {
   if (filedata[inioffset+offsets[i]+2+1] != 0x49) {       // record type: 47-eng,48-pol,49-info
    noffsets[j] = offsets[i];
    words[j] = new char [strlen(&filedata[inioffset+noffsets[j]+2+4+6])+1];
    strcpy(words[j],&filedata[inioffset+noffsets[j]+2+4+6]);
    if ((action == listEnWords) || (action == listPlWords)) {
     cout <<  "words[" << j << "] = " << words[j] << endl;
    }
    j++;
   }
  }
 } else {
  printf("MAP_FAILED\n");
 }

 close(f); 
 delete [] offsets;
 offsets = noffsets;
 wordCount = j;

#ifdef DEBUG
 cout << "wordCount=" << wordCount << endl; 
#endif
}

// Search for word and print it's translation
void cDictionary::read_definition(string word) {
 int w;
 bool found = false;
 string tmpWord;
#ifdef DEBUG
 cout << "wordCount: " << wordCount;
#endif
 // find index of word in words[] table
 for (w=0; w < wordCount; w++ ) {
  tmpWord = words[w];
  replace(&tmpWord,"|","");
  win2iso(&tmpWord);
  // remove_tags(&tmpWord);

  // word found
  if (tmpWord == word) {
#ifdef DEBUG
   cout << "w=" << w << " words[w]=" << tmpWord << endl;
#endif
   found = true;
   break;
  // for words with more than one definitions
  } else if (tmpWord.find(word+"<") == 0) {
#ifdef DEBUG
   cout << "w=" << w << " words[w]=" << tmpWord << endl;
#endif
   found = true;
   break;
  }

  if (!words[w+1]) {
   found = false;
   break;
  }
 }

 if (!found) {
  cout << "Przykro mi, nie znalazem sowa " << word << " w sowniku." << endl;
  exit(1);
 }

 print_definition(w);

 // if there are more than one definitions of word - print them
 while (tmpWord.find(word+"<SUP>") != string::npos) {
  w++;
  print_definition(w);
  tmpWord = words[w+1];
 }
}

// Read and parse configuration file
void read_config(const char *config_file_name) {
 ifstream config_file;
 config_file.open(config_file_name);
 char config_line[256];
 while (!config_file.eof()) {
  config_file.getline(config_line,256); 
  cout << config_line;
 }
 config_file.close();
}

// Show parameters help
void show_help(const char *program_name) {
 cout << "Uruchamiaj:" << endl;
 cout << program_name << " <-le|-lp|-p|--2pl|-e|--2en> [szukane_sowo]" << endl;
}

// Check command line parameters
void check_params(int argc, char *argv[]) {
 if (argc < 2) {
  show_help(argv[0]);
  exit(0);
 }

 const char *options = argv[1];

 if (!strcmp(options,"-p") || !strcmp(options,"--2pl")) {
  strcpy( dict_file_name, angpol_dict_file); 
  action = toPl;
  word = argv[2];
 } else if (!strcmp(options,"-e") || !strcmp(options,"--2en")) {
  strcpy( dict_file_name, polang_dict_file); 
  action = toEn;
  word = argv[2];
 } else if (!strcmp(options,"-le")) {
  strcpy( dict_file_name, angpol_dict_file); 
  action = listEnWords;   
 } else if (!strcmp(options,"-lp")) {
  strcpy( dict_file_name, polang_dict_file); 
  action = listPlWords;   
 } else {
  show_help(argv[0]);
  exit(0);
 }

#ifdef DEBUG
 cout << "Szukane sowo: " << word << endl;
#endif
}

int main(int argc, char *argv[]) {

 cout << "PWN English <-> Polish Dictionary Parser v0.1" << endl;
 
// read_config(config_file_path);
// return 0;
 check_params(argc, argv);    // check command line parameters

 cDictionary dict;
 dict.read_bases();           // read bases,offsets and indexes
 dict.read_words();           // read all words table

 if ((action == toPl) || (action == toEn)) {
  dict.read_definition(word);  // search for word
 }

 return 0;
}
